import json
import numpy as np
from numbers import Number

import rasterio as rio
from rasterio.fill import fillnodata
from rasterio.windows import Window
import riomucho

from scipy.ndimage.filters import maximum_filter, minimum_filter


def pad_window(wnd, pad):
    try:
        wnd = wnd.toranges()
    except AttributeError:
        # rasterio < 1.0, already a tuple
        pass

    return (
        (wnd[0][0] - pad, wnd[0][1] + pad),
        (wnd[1][0] - pad, wnd[1][1] + pad)
    )


def test_rgb(count, nodata, alphafy, outCount):
    if count == 3 and alphafy:
        if not isinstance(nodata, Number):
            raise ValueError('3 band imagery must have a defined nodata value')

        return None, nodata, outCount
    elif alphafy:
        return None, None, count
    else:
        return nodata, nodata, count


def fill_nodata(img, mask, fillBands, maxSearchDistance):
    for b in fillBands:
        img[b - 1] = fillnodata(img[b - 1], mask, maxSearchDistance)

    return img


def runNodataFiller(mask, pad):
    nonZero = np.count_nonzero(mask[pad:-pad, pad:-pad])

    if nonZero == 0 or nonZero == mask[pad:-pad, pad:-pad].size:
        return False
    else:
        return True


def handle_RGB(img, mask):
    return np.concatenate([img, mask.reshape(1, mask.shape[-2], mask.shape[-1])])


def blob_worker(srcs, window, ij, globalArgs):
    pad = globalArgs['max_search_distance'] + 1
    padded = pad_window(window, pad)
    padWindow = Window.from_slices(*padded, boundless=True)
    img = srcs[0].read(boundless=True, window=padWindow)

    if isinstance(globalArgs['selectNodata'], Number):
        mask = srcs[0].read_masks(boundless=True, window=padWindow)[0]
        img = handle_RGB(img, mask)
        alphamask = False
    else:
        mask = img[-1]
        alphamask = True

    if globalArgs['maskThreshold'] is not None and alphamask:
        img[-1] = (np.invert(img[-1] < globalArgs['maskThreshold']).astype(img.dtype)
                   * np.iinfo(img.dtype).max)
        mask = img[-1]

    if runNodataFiller(mask, pad):
        img = fill_nodata(
            img, mask, globalArgs['bands'],
            globalArgs['max_search_distance'])[:, pad: -pad, pad: -pad]

        if globalArgs['nibblemask'] \
                and alphamask is False \
                and 'nodata' in srcs[0].meta:

            img = nibble_filled_mask(
                img,
                srcs[0].meta['nodata'],
                globalArgs['max_search_distance'])

        elif globalArgs['nibblemask'] and alphamask:
            img[-1] = nibble_filled_mask(
                img[-1],
                None,
                globalArgs['max_search_distance'],
                True)

    else:
        img = img[:, pad: -pad, pad: -pad]

    return img


def blob_nodata(
        src_path, dst_path, bidx, max_search_distance, nibblemask,
        creation_options, maskThreshold, workers, alphafy):
    """
    """
    with rio.open(src_path) as src:
        windows = [
            [window, ij] for ij, window in src.block_windows()
        ]

        options = src.meta.copy()
        kwds = src.profile.copy()

        outNodata, selectNodata, outCount = test_rgb(src.count, src.nodata, alphafy, 4)

        options.update(**kwds)
        options.update(**creation_options)
        options.update(count=outCount, nodata=outNodata)

        if bidx:
            try:
                bidx = [int(b) for b in json.loads(bidx)]
            except Exception as e:
                raise e

            if bidx and (len(bidx) == 0 or len(bidx) > src.count):
                raise ValueError(
                    "Bands %s differ from source count of %s" %
                    (', '.join([str(b) for b in bidx]), src.count))
        elif alphafy and src.count == 3:
            bidx = list(src.indexes)
            bidx.append(src.indexes[-1] + 1)
        else:
            bidx = list(src.indexes)

        if maskThreshold is not None:
            maskThreshold = np.iinfo(options['dtype']).max - maskThreshold

    with riomucho.RioMucho(
            [src_path], dst_path, blob_worker,
            windows=windows,
            global_args={
                'max_search_distance': max_search_distance,
                'nibblemask': nibblemask,
                'bands': bidx,
                'maskThreshold': maskThreshold,
                'selectNodata': selectNodata
            },
            options=options,
            mode='manual_read') as rm:

        rm.run(workers)


def nibble_filled_mask(filled, nodataval, max_search_distance, is_mask=False):
    filled = np.array(filled)
    if is_mask:
        filled = minimum_filter(filled, size=(max_search_distance * 2 + 1))
    else:
        nmsk = (filled[0] == nodataval) & (filled[1] == nodataval) & (filled[2] == nodataval)
        nmsk = maximum_filter(nmsk, size=(max_search_distance * 2 + 1))
        mskInds = np.where(nmsk == True)
        filled[:, mskInds[0], mskInds[1]] = nodataval

    return filled


def make_nibbled(src_path, dst_path, nibble):
    with rio.open(src_path, 'r') as src:
        img = src.read(masked=False)
        kwargs = src.meta
        if 'nodata' in src.meta:
            nodataval = src.meta['nodata']
        else:
            nodataval = 0.0

    kwargs.update(compress='lzw')

    nibbled = nibble_filled_mask(img, nodataval, nibble)

    with rio.open(dst_path, 'w', **kwargs) as dst:
        for i, arr in enumerate(nibbled, 1):
            dst.write(arr, i)



from itertools import repeat
from multiprocessing import cpu_count, Pool
import sys
import zlib
try:
    from itertools import izip
except ImportError:
    izip = zip  # py3 use builtin zip

import numpy
import rasterio


# We're compelled to use global variables for the source dataset and masking
# algorithm function. The worker initialization and finalization functions
# make sure that the globals of a worker process are set up and cleaned up
# properly.

src_dataset = None
mask_function = None


def init_worker(path, func):
    global mask_function, src_dataset
    mask_function = func
    src_dataset = rasterio.open(path)


def finalize_worker():
    global mask_function, src_dataset
    src_dataset.close()
    mask_function = None
    src_dataset = None


# The following function is executed by worker processes.

def compute_window_mask(args):
    """Execute the given function with keyword arguments to compute a
    valid data mask.

    Returns the window and mask as deflated bytes.
    """
    window, nodata, extra_args = args
    global mask_function, src_dataset

    padding = int(extra_args.get('padding', 0))

    if padding:
        start, stop = zip(*window)
        start = [v - 10 for v in start]
        stop = [v + 10 for v in stop]
        read_window = zip(*[start, stop])
    else:
        read_window = window

    source = src_dataset.read(window=read_window, boundless=True)
    result = mask_function(source, nodata, **extra_args)

    if padding:
        result = result[:, padding:-padding, padding:-padding]

    return window, zlib.compress(result.tobytes())


def all_valid(arr, nodata, **kwargs):
    """Return an all-valid mask of the same shape and type as the
    given array"""
    return 255 * numpy.ones_like(arr[0])


class NodataPoolMan:
    """Nodata processing pool manager

    This class encapsulates the execution of nodata algorithms on
    windows of a dataset.
    """

    def __init__(self, input_path, func, nodata, num_workers=None,
            max_tasks=100):
        """Create a pool of workers to process window masks"""
        self.input_path = input_path
        self.func = func
        self.nodata = nodata

        # Peek in the source file for metadata. We could even get the
        # nodata value from here in some cases.
        with rasterio.open(input_path) as src:
            self.dtype = src.dtypes[0]

        self.pool = Pool(
            num_workers or cpu_count()-1, init_worker, (input_path, func),
            max_tasks)

    def mask(self, windows, **kwargs):
        """Iterate over windows and compute mask arrays.

        The keyword arguments will be passed as keyword arguments to the
        manager's mask algorithm function.

        Yields window, ndarray pairs.
        """
        iterargs = izip(windows, repeat(self.nodata), repeat(kwargs))
        for out_window, data in self.pool.imap_unordered(
                compute_window_mask, iterargs):

            out_data = numpy.fromstring(
                        zlib.decompress(data), self.dtype).reshape(
                            [int(x) for x in rasterio.windows.shape(out_window)])

            yield out_window, out_data



from __future__ import print_function
import click

from rasterio.rio.options import creation_options

from nodata.blob import blob_nodata


@click.group()
def cli():
    """There's no data like nodata!"""
    pass


@click.command(
    short_help="Blob + expand valid data area by (inter|extra)polation into"
               "nodata areas")
@click.argument('src_path', type=click.Path(exists=True))
@click.argument('dst_path', type=click.Path(exists=False))
@click.option('--bidx', '-b', default=None,
    help="Bands to blob [default = all]")
@click.option('--max-search-distance', '-m', default=4,
    help="Maximum blobbing radius [default = 4]")
@click.option('--nibblemask', '-n', default=False, is_flag=True,
    help="Nibble blobbed nodata areas [default=False]")
@creation_options
@click.option('--mask-threshold', '-d', default=None, type=int,
    help="Alpha pixel threshold upon which to regard data as masked "
         "(ie, for lossy you'd want an aggressive threshold of 0) "
         "[default=None]")
@click.option('--jobs', '-j', default=4, type=int,
    help="Number of workers for multiprocessing [default=4]")
@click.option('--alphafy', '-a', is_flag=True,
    help='If a RGB raster is found, blob + add alpha band where nodata is')
def blob(src_path, dst_path, bidx, max_search_distance, nibblemask,
        creation_options, mask_threshold, jobs, alphafy):
    """"""
    args = (src_path, dst_path, bidx, max_search_distance, nibblemask,
            creation_options, mask_threshold, jobs, alphafy)
    blob_nodata(
        src_path, dst_path, bidx, max_search_distance, nibblemask,
        creation_options, mask_threshold, jobs, alphafy)


cli.add_command(blob)



import numpy as np
import skimage.measure as measure
import scipy.ndimage.measurements as sci_meas
from skimage.segmentation import slic
from scipy.ndimage.morphology import binary_fill_holes


def _diff_nodata(image, ndv):
    return np.sum([np.abs(im.astype(float) - n) for im, n in zip(image, ndv)], axis=0)


def _hacky_make_image(labeled_img, u_labels, measures, m_key, dtype=np.int16):
    out = np.zeros(labeled_img.shape, dtype=dtype).ravel()
    if m_key is None:
        measures = [{"key": m} for m in measures]
        m_key = "key"

    def value_grab(a, b):
        out[b] = measures[a[0] - 1][m_key]
        return None

    sci_meas.labeled_comprehension(labeled_img, labeled_img, u_labels, value_grab, float, 0, pass_positions=True)

    return out.reshape(labeled_img.shape)


def all_valid(data, ndv, threshold=0):
    """Test if all the data in the entire array are valid data
    within the specified threshold"""
    diff = _diff_nodata(data, ndv)
    return np.all(diff > (threshold * data.shape[0]))


def _edges(arr):
    left = arr[0, :]
    right = arr[-1, :]
    top = arr[:, 0]
    bottom = arr[:, -1]
    edges = np.concatenate([left, right, top, bottom])
    return edges


def all_valid_edges(data, ndv, threshold=0):
    """Test if all the data on the edges are valid data
    within the specified threshold"""
    diff = _diff_nodata(data, ndv)
    edges = _edges(diff)
    return np.all(edges > (threshold * data.shape[0]))


def simple_mask(data, ndv):
    '''Exact nodata masking'''
    depth, rows, cols = data.shape
    nd = np.iinfo(data.dtype).max
    alpha = np.invert(np.all(np.dstack(data) == ndv, axis=2)).astype(data.dtype) * nd
    return alpha


def slic_mask(arr, nodata, n_clusters=50, threshold=5, debug=False):
    """
    Uses @dnomadb algorithm, roughly:
        - cluster image using SLIC (k-means)
        - pull out contiguous regions and find aggregate stats
        - select regions that are likely ndv
        - fill inclusions
    """
    assert arr.shape[0] == len(nodata)
    near_nodata = _diff_nodata(arr, nodata)
    clusters = slic(near_nodata, n_clusters)
    labeled = measure.label(clusters) + 1
    measures = measure.regionprops(labeled, intensity_image=near_nodata, cache=True)
    mean_intensity = _hacky_make_image(labeled, np.unique(labeled), measures, 'mean_intensity')
    mask = binary_fill_holes(np.invert(binary_fill_holes(mean_intensity >= threshold)))
    nd = np.iinfo(arr.dtype).max
    mask = np.invert(mask) * nd
    if debug:
        d = dict([(m.label, m.mean_intensity) for m in measures])
        return mask.astype('uint8'), labeled.astype('uint32'), mean_intensity.astype('uint32'), d
    else:
        return mask.astype('uint8')
